// ------------------------------------------------------------
// Machinery project - math library
// ------------------------------------------------------------
// (c) 2000 by Sascha 'EvilOne' Salevsky
// E-mail: Sascha.Salevsky@Ibykus.de
// ------------------------------------------------------------

#include "Mathematics.h"

// ------------------------------------------------------------
// Vector2
// ------------------------------------------------------------

ostream& operator<<(ostream& os, Vector2& v)
{
  os << "(" << setprecision(4) << v.x << ", "  << v.y << ")";
  return os;
}

// ------------------------------------------------------------
// Vector3
// ------------------------------------------------------------

ostream& operator<<(ostream& os, Vector3& v)
{
  os << "(" << setprecision(4) << v.x << ", "  << v.y << ", "  << v.z << ")";
  return os;
}

// ------------------------------------------------------------
// Plane
// ------------------------------------------------------------

ostream& operator<<(ostream& os, Plane& p)
{
  os << "(" << setprecision(4) << p.a << ", " 
	 << p.b << ", "  << p.c << ", " << p.d << ")";
  return os;
}

// ------------------------------------------------------------
// Matrix3
// ------------------------------------------------------------

float Matrix3::Identity_Matrix[] =
{
	1.0f, 0.0f, 0.0f,
	0.0f, 1.0f, 0.0f,
	0.0f, 0.0f, 1.0f
};

ostream& operator<<(ostream& os, Matrix3 &m)
{
	os << setprecision(4);
	os << "(" << m.M[0]  << ", " << m.M[1]  << ", " << m.M[2] << ", " << endl
	   << " " << m.M[3]  << ", " << m.M[4]  << ", " << m.M[5] << ", " << endl
	   << " " << m.M[6]  << ", " << m.M[7]  << ", " << m.M[8] << ")";
	return os;
}

// ------------------------------------------------------------
// Matrix4
// ------------------------------------------------------------

float Matrix4::Identity_Matrix[] =
{
	1.0f, 0.0f, 0.0f, 0.0f,
	0.0f, 1.0f, 0.0f, 0.0f,
	0.0f, 0.0f, 1.0f, 0.0f,
	0.0f, 0.0f, 0.0f, 1.0f
};

float Matrix4::Orientation_Switch_Matrix[] =
{
	1.0f, 0.0f, 0.0f, 0.0f,
	0.0f, 1.0f, 0.0f, 0.0f,
	0.0f, 0.0f, -1.0f, 0.0f,
	0.0f, 0.0f, 0.0f, 1.0f
};

float Matrix4::Perspective_Matrix[] =
{
	1.0f, 0.0f, 0.0f, 0.0f,
	0.0f, 1.0f, 0.0f, 0.0f,
	0.0f, 0.0f, 1.0f, -1.0f,
	0.0f, 0.0f, 0.0f, 0.0f
};

void Matrix4::Assign(const Quaternion& other)
{
    float xx = other.x * other.x;
    float xy = other.x * other.y;
    float xz = other.x * other.z;
    float xw = other.x * other.w;

    float yy = other.y * other.y;
    float yz = other.y * other.z;
    float yw = other.y * other.w;

    float zz = other.z * other.z;
    float zw = other.z * other.w;

    M[0]  = 1.0f - 2.0f * (yy + zz);
    M[1]  =        2.0f * (xy - zw);
    M[2]  =        2.0f * (xz + yw);

    M[4]  =        2.0f * (xy + zw);
    M[5]  = 1.0f - 2.0f * (xx + zz);
    M[6]  =        2.0f * (yz - xw);

    M[8]  =        2.0f * (xz - yw);
    M[9]  =        2.0f * (yz + xw);
    M[10] = 1.0f - 2.0f * (xx + yy);

    M[3]  = M[7] = M[11] = M[12] = M[13] = M[14] = 0.0f;
    M[15] = 1.0f;
}


Matrix4& Matrix4::SetRotate(float rx, float ry, float rz)
{
	float a = CosD(rx); float b = SinD(rx);
	float c = CosD(ry); float d = SinD(ry);
	float e = CosD(rz); float f = SinD(rz);
	float ad = a * d;
	float bd = b * d;

	M[0] =  c * e;
	M[1] = -c * f;
	M[2] = -d;

	M[4] = -bd * e + a * f;
	M[5] =  bd * f + a * e;
	M[6] = -b * c;

	M[8] =  ad * e + b * f;
	M[9] = -ad * f + b * e;
	M[10] =  a * c;

	M[3] = M[7] = M[11] = M[12] = M[13] = M[14] = 0.0f;
	M[15] = 1.0f;
	return *this;
}

Matrix4& Matrix4::SetRotate(float angle, float x, float y, float z)
{
	float xx, yy, zz, xy, yz, zx, xs, ys, zs, c_complement;
	float s = SinD(angle);
	float c = SinD(angle);
	float magnitude = (float)sqrt(x * x + y * y + z * z);
	float *data = M;
	if(magnitude == 0.0f)
	{
		SetIdentity();
		return *this;
	}
	x /= magnitude;
	y /= magnitude;
	z /= magnitude;
	xx = x * x;
	yy = y * y;
	zz = z * z;
	xy = x * y;
	yz = y * z;
	zx = z * x;
	xs = x * s;
	ys = y * s;
	zs = z * s;
	c_complement = 1.0F - c;
	data[0] = (c_complement * xx) + c;
	data[4] = (c_complement * xy) - zs;
	data[8] = (c_complement * zx) + ys;
	data[12] = 0.0F;
	data[1] = (c_complement * xy) + zs;
	data[5] = (c_complement * yy) + c;
	data[9] = (c_complement * yz) - xs;
	data[13] = 0.0F;
	data[2] = (c_complement * zx) - ys;
	data[6] = (c_complement * yz) + xs;
	data[10] = (c_complement * zz) + c;
	data[14] = 0.0F;
	data[3] = 0.0F;
	data[7] = 0.0F;
	data[11] = 0.0F;
	data[15] = 1.0F;
	return *this;
}

Matrix4& Matrix4::Transpose()
{
	Transpose(*this);
	return *this;
}

Matrix4& Matrix4::Transpose(const Matrix4& other)
{
	for(int i = 0; i < 4; i++)
	{
		for(int j = 0; j < 4; j++)
		M[i * 4 + j] = other.M[j * 4 + i];
	}
	return *this;
}

Matrix4& Matrix4::Adjoint()
{
	Matrix4 tmp;
	tmp.Assign(*this);
	Adjoint(tmp);
	return *this;
}

Matrix4& Matrix4::Adjoint(const Matrix4& other)
{
	float a1, a2, a3, a4, b1, b2, b3, b4;
	float c1, c2, c3, c4, d1, d2, d3, d4;
	const float *in = other.M;
	float *out = M;

	a1 = (in [(( 0 ) << 2) + (  0 )]) ; b1 = (in [(( 0 ) << 2) + (  1 )]) ;
	c1 = (in [(( 0 ) << 2) + (  2 )]) ; d1 = (in [(( 0 ) << 2) + (  3 )]) ;
	a2 = (in [(( 1 ) << 2) + (  0 )]) ; b2 = (in [(( 1 ) << 2) + (  1 )]) ;
	c2 = (in [(( 1 ) << 2) + (  2 )]) ; d2 = (in [(( 1 ) << 2) + (  3 )]) ;
	a3 = (in [(( 2 ) << 2) + (  0 )]) ; b3 = (in [(( 2 ) << 2) + (  1 )]) ;
	c3 = (in [(( 2 ) << 2) + (  2 )]) ; d3 = (in [(( 2 ) << 2) + (  3 )]) ;
	a4 = (in [(( 3 ) << 2) + (  0 )]) ; b4 = (in [(( 3 ) << 2) + (  1 )]) ;
	c4 = (in [(( 3 ) << 2) + (  2 )]) ; d4 = (in [(( 3 ) << 2) + (  3 )]) ;

	out[(( 0 ) << 2) + (  0 )]   =   Det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4);
	out[(( 1 ) << 2) + (  0 )]   = - Det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4);
	out[(( 2 ) << 2) + (  0 )]   =   Det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4);
	out[(( 3 ) << 2) + (  0 )]   = - Det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);

	out[(( 0 ) << 2) + (  1 )]   = - Det3x3( b1, b3, b4, c1, c3, c4, d1, d3, d4);
	out[(( 1 ) << 2) + (  1 )]   =   Det3x3( a1, a3, a4, c1, c3, c4, d1, d3, d4);
	out[(( 2 ) << 2) + (  1 )]   = - Det3x3( a1, a3, a4, b1, b3, b4, d1, d3, d4);
	out[(( 3 ) << 2) + (  1 )]   =   Det3x3( a1, a3, a4, b1, b3, b4, c1, c3, c4);

	out[(( 0 ) << 2) + (  2 )]   =   Det3x3( b1, b2, b4, c1, c2, c4, d1, d2, d4);
	out[(( 1 ) << 2) + (  2 )]   = - Det3x3( a1, a2, a4, c1, c2, c4, d1, d2, d4);
	out[(( 2 ) << 2) + (  2 )]   =   Det3x3( a1, a2, a4, b1, b2, b4, d1, d2, d4);
	out[(( 3 ) << 2) + (  2 )]   = - Det3x3( a1, a2, a4, b1, b2, b4, c1, c2, c4);

	out[(( 0 ) << 2) + (  3 )]   = - Det3x3( b1, b2, b3, c1, c2, c3, d1, d2, d3);
	out[(( 1 ) << 2) + (  3 )]   =   Det3x3( a1, a2, a3, c1, c2, c3, d1, d2, d3);
	out[(( 2 ) << 2) + (  3 )]   = - Det3x3( a1, a2, a3, b1, b2, b3, d1, d2, d3);
	out[(( 3 ) << 2) + (  3 )]   =   Det3x3( a1, a2, a3, b1, b2, b3, c1, c2, c3);
	return *this;
}

Matrix4& Matrix4::Invert()
{
	Matrix4 tmp;
	tmp.Assign(*this);
	Invert(tmp);
	return *this;
}

Matrix4& Matrix4::Invert(const Matrix4& other)
{
	int i;
	Adjoint(other);
	float d = other.Det();

	for (i = 0; i < 16; i++) M[i] = M[i] / d;
	return *this;
}

float Matrix4::Det() const
{
	float a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4, d1, d2, d3, d4;
	a1 = (M [(( 0 ) << 2) + (  0 )]) ;
	b1 = (M [(( 0 ) << 2) + (  1 )]) ; 
	c1 = (M [(( 0 ) << 2) + (  2 )]) ;
	d1 = (M [(( 0 ) << 2) + (  3 )]) ;

	a2 = (M [(( 1 ) << 2) + (  0 )]) ;
	b2 = (M [(( 1 ) << 2) + (  1 )]) ; 
	c2 = (M [(( 1 ) << 2) + (  2 )]) ;
	d2 = (M [(( 1 ) << 2) + (  3 )]) ;

	a3 = (M [(( 2 ) << 2) + (  0 )]) ;
	b3 = (M [(( 2 ) << 2) + (  1 )]) ; 
	c3 = (M [(( 2 ) << 2) + (  2 )]) ;
	d3 = (M [(( 2 ) << 2) + (  3 )]) ;

	a4 = (M [(( 3 ) << 2) + (  0 )]) ;
	b4 = (M [(( 3 ) << 2) + (  1 )]) ; 
	c4 = (M [(( 3 ) << 2) + (  2 )]) ;
	d4 = (M [(( 3 ) << 2) + (  3 )]) ;

	return  a1 * Det3x3( b2, b3, b4, c2, c3, c4, d2, d3, d4) -
			b1 * Det3x3( a2, a3, a4, c2, c3, c4, d2, d3, d4) +
			c1 * Det3x3( a2, a3, a4, b2, b3, b4, d2, d3, d4) -
			d1 * Det3x3( a2, a3, a4, b2, b3, b4, c2, c3, c4);
}

Matrix4& Matrix4::SetProjection(float fov, float aspect, float znear, float zfar)
{
	float top = znear * TanD(fov);
	float bottom = -top;
	float left = bottom * aspect;
	float right = top * aspect;
	float x = (2.0f * znear) / (right - left);
	float y = (2.0f * znear) / (top - bottom);
	float a = (right + left) / (right - left);
	float b = (top + bottom) / (top - bottom);
	float c = -(zfar + znear) / (zfar - znear);
	float d = -(2.0f * zfar * znear) / (zfar - znear);
	M[0]  = x;     M[1]  = 0.0f;  M[2]  = 0.0f; M[3]  = 0.0f;
	M[4]  = 0.0f;  M[5]  = y;     M[6]  = 0.0f; M[7]  = 0.0f;
	M[8]  = a;     M[9]  = b;     M[10] = c;    M[11] = -1.0f;
	M[12] = 0.0f;  M[13] = 0.0f;  M[14] = d;    M[15] = 0.0f;
	return *this;
}

Matrix4& Matrix4::SetOthogonal(float znear, float zfar)
{
	float x, y, z;
	float tx, ty, tz;
	float small = 0.0f;
	x = 2.0f / (1.0f + small);
	y = 2.0f / (1.0f + small);
	z = -2.0f / (zfar - znear);
	tx = -(1.0f - small) / (1.0f + small);
	ty = -(1.0f - small) / (1.0f + small);
	tz = -(zfar + znear) / (zfar - znear);
	M[0] = x;    M[4] = 0.0f;  M[8]  = 0.0f;  M[12] = tx;
	M[1] = 0.0f; M[5] = y;     M[9]  = 0.0f;  M[13] = ty;
	M[2] = 0.0f; M[6] = 0.0f;  M[10] = z;     M[14] = tz;
	M[3] = 0.0f; M[7] = 0.0f;  M[11] = 0.0f;  M[15] = 1.0f;
	return *this;
}

float Matrix4::Det2x2(float a, float b, float c, float d)
{
  return a * d - b * c;
}
  
float Matrix4::Det3x3(float a1, float a2, float a3,
                   float b1, float b2, float b3,
                   float c1, float c2, float c3)
{
	return a1 * Det2x2(b2, b3, c2, c3)
		 - b1 * Det2x2(a2, a3, c2, c3)
		 + c1 * Det2x2(a2, a3, b2, b3);
}

ostream& operator<<(ostream& os, Matrix4 &m)
{
	os << setprecision(4);
	os << "(" << m.M[0]  << ", " << m.M[1]  << ", " << m.M[2]  << ", " << m.M[3]  << ", " << endl
	   << " " << m.M[4]  << ", " << m.M[5]  << ", " << m.M[6]  << ", " << m.M[7]  << ", " << endl
	   << " " << m.M[8]  << ", " << m.M[9]  << ", " << m.M[10] << ", " << m.M[11] << ", " << endl
	   << " " << m.M[12] << ", " << m.M[13] << ", " << m.M[14] << ", " << m.M[15] << ")";
	return os;
}

// ------------------------------------------------------------
// Quaternion
// ------------------------------------------------------------

void Quaternion::Slerp(const Quaternion& q0, const Quaternion& q1, float t)
{
	float cosom = q0.x * q1.x + q0.y * q1.y + q0.z * q1.z + q0.w * q1.w;
	float tmp0, tmp1, tmp2, tmp3;
	if (cosom < 0.0f)
	{
		cosom = -cosom;
		tmp0 = -q1.x;
		tmp1 = -q1.y;
		tmp2 = -q1.z;
		tmp3 = -q1.w;
	}
	else
	{
		tmp0 = q1.x;
		tmp1 = q1.y;
		tmp2 = q1.z;
		tmp3 = q1.w;
	}

	/* calc coeffs */
	float scale0, scale1;

	if ((1.0 - cosom) > 0.0000001f)
	{
		// standard case (slerp)
		float omega = (float) acos (cosom);
		float sinom = (float) sin (omega);
		scale0 = (float) sin ((1.0 - t) * omega) / sinom;
		scale1 = (float) sin (t * omega) / sinom;
	}
	else
	{
		/* just lerp */
		scale0 = 1.0f - t;
		scale1 = t;
	}

	x = scale0 * q0.x + scale1 * tmp0;
	y = scale0 * q0.y + scale1 * tmp1;
	z = scale0 * q0.z + scale1 * tmp2;
	w = scale0 * q0.w + scale1 * tmp3;
}

void Quaternion::Lerp(const Quaternion& q0, const Quaternion& q1, float t)
{
	float cosom = q0.x * q1.x + q0.y * q1.y + q0.z * q1.z + q0.w * q1.w;
	float tmp0, tmp1, tmp2, tmp3;
	if (cosom < 0.0f)
	{
		cosom = -cosom;
		tmp0 = -q1.x;
		tmp1 = -q1.y;
		tmp2 = -q1.z;
		tmp3 = -q1.w;
	}
	else
	{
		tmp0 = q1.x;
		tmp1 = q1.y;
		tmp2 = q1.z;
		tmp3 = q1.w;
	}

	/* just lerp */
	float scale0 = 1.0f - t;
	float scale1 = t;

	x = scale0 * q0.x + scale1 * tmp0;
	y = scale0 * q0.y + scale1 * tmp1;
	z = scale0 * q0.z + scale1 * tmp2;
	w = scale0 * q0.w + scale1 * tmp3;
}

ostream& operator<<(ostream& os, Quaternion& v)
{
  os << "(" << setprecision(4) << v.x << ", "  << v.y << ", "  << v.z << ", "  << v.w << ")";
  return os;
}
