// ------------------------------------------------------------
// Machinery project - math library
// ------------------------------------------------------------
// (c) 2000 by Sascha 'EvilOne' Salevsky
// E-mail: Sascha.Salevsky@Ibykus.de
// ------------------------------------------------------------
//
// Erm, all angles are in degree(like in OGL), I hate this
// fuckin' radians...
//
// Hmmm, some optimizations can be done, mostly in Matrix4,
// but for the first version I want working code, not 100%
// optimal one, tuning comes later...
//
// Currently some camera-stuff is missing in Matrix4, by
// the way, this comes later ... at the time I need it ;-)
// I think an higher level camera class may be better?!?
//
// should I include defines like #define FTYPE float
// and use this abstract type, so that it can also be used
// with doubles?????
//
// Hmmmmm, I think about making this all a template-library...
//
//
// Hakuna Matata, EvilOne
// Bite me... hehe
// ------------------------------------------------------------

#ifndef _mathematics_h_
#define _mathematics_h_

#include <math.h>
#include <stdlib.h> 
#include <iostream.h>
#include <iomanip.h>

// ------------------------------------------------------------
// constants
// ------------------------------------------------------------

// PI & Co.

const float M_PI				= 3.14159265358979323846f;
const float M_DEG_TO_RAD		= 0.01745329251994329547f;
const float M_RAD_TO_DEG		= 57.29577951308232286465f;
const float M_PI_DIV_2			= 1.570796326794896558f;
const float M_PI_DIV_4			= 0.785398163397448279f;

// tolerance values

const float M_LOW_TOLERANCE		= 0.000001f;
const float M_TOLERANCE			= 0.000010f;
const float M_HIGH_TOLERANCE	= 0.000100f;

// ------------------------------------------------------------
// basic functions
// ------------------------------------------------------------

inline float Clamp(float f, float min, float max)
{
	return min > f ? min : (max > f ? f : max);
}

inline float Lerp(float f, float from, float to)
{
	return from + (to - from) * f;
}

inline float Min(float f1, float f2)
{
	return f1 < f2 ? f1 : f2;
}

inline float Max(float f1, float f2)
{
	return f1 > f2 ? f1 : f2;
}

inline float SinD(float f)
{
	return (float)sin(f * M_DEG_TO_RAD);
}

inline float CosD(float f)
{
	return (float)cos(f * M_DEG_TO_RAD);
}

inline float TanD(float f)
{
	return (float)tan(f * M_DEG_TO_RAD);
}

inline float AtanD(float f)
{
	return (float)atan(f * M_DEG_TO_RAD);
}

inline float Degree(float f)
{
	return f * M_DEG_TO_RAD;
}

inline float Radians(float f)
{
	return f * M_RAD_TO_DEG;
}

inline bool IsZeroL(float f)
{
	return fabs(f) > M_LOW_TOLERANCE;
}

inline bool IsZero(float f)
{
	return fabs(f) > M_TOLERANCE;
}

inline bool IsZeroH(float f)
{
	return fabs(f) > M_HIGH_TOLERANCE;
}

inline float Rand(float max)
{
	return (float(rand()) / float(RAND_MAX)) * max;
}

inline float Rand(float min, float max)
{
	return Rand(max - min) + min; 
}

// ------------------------------------------------------------
// Linear algebra classes
// ------------------------------------------------------------

class Vector2;
class Matrix3;
class Vector3;
class Plane;
class Matrix4;
class Quaternion;

// ------------------------------------------------------------
// Vector2
// ------------------------------------------------------------

class Vector2
{
public:

	// Members

	float x, y;

public:

	// Constructors

	Vector2();
	Vector2(const float* other);
	Vector2(float _x, float _y);
	Vector2(const Vector2& other);

	// Assign

	Vector2& operator=(const Vector2& other);
	void Assign(float _x, float _y);
	void Assign(const Vector2& other);
	void Assign(float* other);

	// Swap

	void Swap(Vector2& other);

	// Casting

	operator float*();
	operator const float*() const;

	// Math

	// this += other
	void Add(const Vector2& other);
	// this = v1 + v2
	void Add(const Vector2& v1, const Vector2& v2);
	// this -= other
	void Sub(const Vector2& other);
	// this = v1 - v2
	void Sub(const Vector2& v1, const Vector2& v2);
	// this *= f
	void Mul(float f);
	// this = other * f
	void Mul(const Vector2& other, float f);
	// this /= f
	void Div(float f);
	// this = other / f
	void Div(const Vector2& other, float f);
	// this = -this
	void Invert();
	// this = -other
	void Invert(const Vector2& other);

	// Linear combination

	// this += other * s
	void Combine(const Vector2& other, float s);
	// this = this * s + other * t
	void Combine(float s, const Vector2& other, float t);
	// this = v1 * s + v2 * t
	void Combine(const Vector2& v1, float s, const Vector2& v2, float t);

	// operators

	Vector2& operator+=(const Vector2& other);
	Vector2& operator-=(const Vector2& other);
	Vector2& operator*=(float f);
	Vector2& operator/=(float f);
	Vector2& operator+=(float f);
	Vector2& operator-=(float f);
	Vector2& operator-();
	Vector2 operator+(const Vector2& other) const;
	Vector2 operator-(const Vector2& other) const;
	Vector2 operator*(float f) const;
	Vector2 operator/(float f) const;
	Vector2 operator+(float f) const;
	Vector2 operator-(float f) const;

	// Linear Algebra

	float Length() const;
	float LengthSqr() const;
	float Distance(const Vector2& other) const;
	float DistanceSqr(const Vector2& other) const;
	void Normalize();
	void Normalize(const Vector2& other);

	void Rotate(float angle);
	void Rotate(Vector2 other, float angle);

	// Others

	void Clamp(float min, float max);
	// this = v1 + (v2 - v1) * s
	void Lerp(const Vector2& v1, const Vector2& v2, float s);

	// Actions with Matrix3

	void Transform(const Matrix3& mat, const Vector2& other);
	void Transform(const Matrix3& mat);

	// friend operators

	friend ostream& operator<<(ostream& os, Vector2& v);
};

// ------------------------------------------------------------
// Matrix3
// ------------------------------------------------------------

class Matrix3
{
public:

	// members

	union
	{
		float M[9];
		float m[3][3];
	};

	static float Identity_Matrix[];

public:

	// Constructors

	Matrix3();
	Matrix3(const float* other);
	Matrix3(const Matrix3& other);

	// Assignment

	Matrix3& operator=(const Matrix3& other);
	void Assign(const float* other);
	void Assign(const Matrix3& other);

	// Casting

	operator float*();
	operator const float*() const;

	// Simple methods

	Matrix3& SetIdentity();
	Matrix3& SetZero();

	// this *= other
	Matrix3& Mul(const Matrix3& other);
	// this = m1 * m2
	Matrix3& Mul(const Matrix3& m1, const Matrix3& m2);
	// this *= f
	Matrix3& Mul(float f);
	// this = other * f
	Matrix3& Mul(const Matrix3& other, float f);
	// this /= f
	Matrix3& Div(float f);
	// this = other / f
	Matrix3& Div(const Matrix3& other, float f);

	// Operators

	Matrix3& operator*=(const Matrix3& other);
	Matrix3 operator*(const Matrix3& other) const;

	// Transformation methods

	Matrix3& Translate(float tx, float ty);
	Matrix3& SetTranslate(float tx, float ty);
	Matrix3& Rotate(float r);
	Matrix3& SetRotate(float r);
	Matrix3& Scale(float sx, float sy);
	Matrix3& SetScale(float sx, float sy);

	// friends

	friend ostream& operator<<(ostream& os, Matrix3 &m);
};

// ------------------------------------------------------------
// Vector3
// ------------------------------------------------------------

class Vector3
{
public:

	// Member variables

	float x, y, z;

public:

	// Constructors

	Vector3();
	Vector3(float _x, float _y, float _z);
	Vector3(Vector3& other);
	Vector3(const float* other);

	// assignemt

	Vector3& operator=(const Vector3& other);
	void Assign(float _x, float _y, float _z);
	void Assign(const Vector3& other);
	void Assign(float* other);

	// Swap

	void Swap(Vector3& other);

	// Casting

	operator float*();
	operator const float*() const;

	// Math

	// this += other
	void Add(const Vector3& other);
	// this = v1 + v2
	void Add(const Vector3& v1, const Vector3& v2);
	// this -= other
	void Sub(const Vector3& other);
	// this = v1 - v2
	void Sub(const Vector3& v1, const Vector3& v2);
	// this *= f
	void Mul(float f);
	// this = other * f
	void Mul(const Vector3& other, float f);
	// this /= f
	void Div(float f);
	// this = other / f
	void Div(const Vector3& other, float f);
	// this = -this
	void Invert();
	// this = -other
	void Invert(const Vector3& other);

	// Linear combination

	// this += other * s
	void Combine(const Vector3& other, float s);
	// this = this * s + other * t
	void Combine(float s, const Vector3& other, float t);
	// this = v1 * s + v2 * t
	void Combine(const Vector3& v1, float s, const Vector3& v2, float t);

	// EvilOne's beloved overloads

	Vector3& operator+=(const Vector3& other);
	Vector3& operator-=(const Vector3& other);
	Vector3& operator*=(float f);
	Vector3& operator/=(float f);
	Vector3& operator+=(float f);
	Vector3& operator-=(float f);
	Vector3& operator-();
	Vector3 operator+(const Vector3& other) const;
	Vector3 operator-(const Vector3& other) const;
	Vector3 operator*(float f) const;
	Vector3 operator/(float f) const;
	Vector3 operator+(float f) const;
	Vector3 operator-(float f) const;

	// Linear Algebra

	float Dot(const Vector3& v) const;
	// this = v1 x v2
	void Cross(const Vector3& v);
	void Cross(const Vector3& v1, const Vector3& v2);
	float Length() const;
	float LengthSqr() const;
	float Distance(const Vector3& other) const;
	float DistanceSqr(const Vector3& other) const;
	void Normalize();
	void Normalize(const Vector3& other);

	// Other usefull funtions

	void Clamp(float min, float max);
	// this = v1 + (v2 - v1) * s
	void Lerp(const Vector3& v1, const Vector3& v2, float s);

	// Operations with a plane

	void Intersect(const Vector3& start, float s, const Vector3& end, float e);
	void Intersect(const Plane& p, const Vector3& start, const Vector3& end);
	void Reflect(const Plane& p, const Vector3& v);

	// Operations with Matrix

	void Transform(const Matrix4& mat, const Vector3& other);
	void Transform(const Matrix4& mat);
	void TransformTranspose(const Matrix4& mat, const Vector3& other);
	void TransformTranspose(const Matrix4& mat);

	// friend operators

	friend ostream& operator<<(ostream& os, Vector3& v);
};

// ------------------------------------------------------------
// Plane
// ------------------------------------------------------------

class Plane
{
public:

	// members 

	union
	{
		struct
		{
			float a, b, c, d;
		};

		struct
		{
			Vector3 n;
			float d;
		};
	};

public:

	// Constructors

	Plane();
	Plane(const Vector3& _n, float _d);
	Plane(float _a, float _b, float _c, float _d);
	Plane(float* other);
	Plane(const Plane& other);

	// Assignment

	Plane& operator=(const Plane& other);
	void Assign(const Vector3& _n, float _d);
	void Assign(float _a, float _b, float _c, float _d);
	void Assign(float* other);
	void Assign(const Plane& other);

	// Functions

	float Distance(const Vector3& v) const;

	// Interaction with matrix

	void Transform(const Matrix4 &mat);
	void Transform(const Matrix4 &mat, const Plane& other);

	// friend operators

	friend ostream& operator<<(ostream& os, Plane& p);
};

// ------------------------------------------------------------
// Matrix4
// ------------------------------------------------------------

class Matrix4
{
public:

	// members

	union
	{
		float M[16];
		float m[4][4];
	};

	static float Identity_Matrix[];
	static float Orientation_Switch_Matrix[];
	static float Perspective_Matrix[];

public:

	// Constructors

	Matrix4();
	Matrix4(const float* other);
	Matrix4(const Matrix4& other);
	Matrix4(const Quaternion& other);

	// Assignment

	Matrix4& operator=(const Matrix4& other);
	Matrix4& operator=(const Quaternion& other);
	void Assign(const Quaternion& other);
	void Assign(const float* other);
	void Assign(const Matrix4& other);

	// Casting

	operator float*();
	operator const float*() const;

	// Simple methods

	Matrix4& SetIdentity();
	Matrix4& SetZero();
	Matrix4& SetPerspective();
	Matrix4& SetSwitchOrientation();

	// this *= other
	Matrix4& Mul(const Matrix4& other);
	// this = m1 * m2
	Matrix4& Mul(const Matrix4& m1, const Matrix4& m2);
	// this *= f
	Matrix4& Mul(float f);
	// this = other * f
	Matrix4& Mul(const Matrix4& other, float f);
	// this /= f
	Matrix4& Div(float f);
	// this = other / f
	Matrix4& Div(const Matrix4& other, float f);

	// Operators

	Matrix4& operator*=(const Matrix4& other);
	Matrix4 operator*(const Matrix4& other) const;

	// Transformation methods

	Matrix4& SetTranslate(float tx, float ty, float tz);
	Matrix4& Translate(float tx, float ty, float tz);
	Matrix4& SetScale(float sx, float sy, float sz);
	Matrix4& Scale(float sx, float sy, float sz);

	// rotation around three euler-angles

	Matrix4& SetRotate(const Vector3& r);
	Matrix4& SetRotate(float rx, float ry, float rz);
	Matrix4& Rotate(const Vector3& r);
	Matrix4& Rotate(float rx, float ry, float rz);

	// rotation euler-angle around axis
	
	Matrix4& SetRotate(float angle, const Vector3& r);
	Matrix4& SetRotate(float angle, float x, float y, float z);
	Matrix4& Rotate(float angle, const Vector3& r);
	Matrix4& Rotate(float angle, float x, float y, float z);

	// Invert/Transpose

	Matrix4& Adjoint();
	Matrix4& Adjoint(const Matrix4& other);

	Matrix4& Transpose();
	Matrix4& Transpose(const Matrix4& other);

	Matrix4& Invert();
	Matrix4& Invert(const Matrix4& other);

	float Det() const;

	// Perpsective

	Matrix4& SetProjection(float fov, float aspect, float znear, float zfar);
	Matrix4& SetOthogonal(float znear, float zfar);

	// static

	static float Det2x2(float a, float b, float c, float d);
	static float Det3x3(float a1, float a2, float a3,
						float b1, float b2, float b3,
						float c1, float c2, float c3);

	// friends

	friend ostream& operator<<(ostream& os, Matrix4 &m);
};

// ------------------------------------------------------------
// Quaternion
// ------------------------------------------------------------

class Quaternion
{
public:

	// Members

	float x, y, z, w;

public:

	// Constructors

	Quaternion();
	Quaternion(const Quaternion& other);
	Quaternion(const float* other);
	Quaternion(float _x, float _y, float _z, float _w);

	// Assignment

	Quaternion& operator=(const Quaternion& other);
	void Assign(const Quaternion& other);
	void Assign(const float* other);
	void Assign(float _x, float _y, float _z, float _w);

	// Simple math

	void Invert();
	void Invert(const Quaternion& other);
	Quaternion& operator-();

	Quaternion& Mul(const Quaternion& other);
	Quaternion& Mul(const Quaternion& q1, const Quaternion& q2);
	Quaternion operator*(const Quaternion& other) const;
	Quaternion& operator*=(const Quaternion& other);

	float Length() const;
	float LengthSqr() const;

	void Normalize();
	void Normalize(const Quaternion& other);

	// Rotation around euler-angle and axis

	Quaternion& SetRotate(float angle, const Vector3& v);
	Quaternion& Rotate(float angle, const Vector3& v);
	Quaternion& SetRotate(float angle, float _x, float _y, float _z);
	Quaternion& Rotate(float angle, float _x, float _y, float _z);

	// rotation around three euler-angles

	Quaternion& SetRotate(const Vector3& v);
	Quaternion& Rotate(const Vector3& v);
	Quaternion& SetRotate(float rx, float ry, float rz);
	Quaternion& Rotate(float rx, float ry, float rz);

	// Interpolation

	void Lerp(const Quaternion& q0, const Quaternion& q1, float t);
	void Slerp(const Quaternion& q0, const Quaternion& q1, float t);

	// friend operators

	friend ostream& operator<<(ostream& os, Quaternion& v);
};

// ------------------------------------------------------------
// Include implementations
// ------------------------------------------------------------

#include "Vector2.h"
#include "Matrix3.h"
#include "Vector3.h"
#include "Plane.h"
#include "Matrix4.h"
#include "Quaternion.h"

#endif
